"use strict";

const ee = require("./eventemitter");
const eventNamespace = "tgam-photo-slider";
const reqTimeout = require("./request-timeout.js");

const throttle = require("./throttle.js");
throttle("resize", "windowResize.throttled", window);

// comonpnent photo slider
const slideHasInfoClass = "c-photo-slider--has-info";
const slideHasNavClass = "c-photo-slider--has-nav";

const slideListClass = "c-photo-slider__list";
const slideItemClass = "c-photo-slider__item";

const slideNavLeftClass = "c-photo-slider__nav--left";
const slideNavRightClass = "c-photo-slider__nav--right";

// component article media image
const slideMediaClass = "c-article-media__image";
const slideMediaImageFullFitClass = "c-article-media__image--full-fit";

const slideMediaLinkClass = "c-article-media__link";
const slideMediaInfoClass = "c-article-media__info";

// component image
const slideImageClass = "c-image";

module.exports = {

  init: function(rootElement, slideStart = 0) {
    if (!rootElement) return;

    this.rootElement = rootElement;
    this.slideActive = slideStart; // current slide shown
    this.swipeThreshold = 0.2; // slide movement above this width percentage counts as a swipe
    this.slideSpeed = 200; // speed when animating between slides
    this.touchSpeed = 20; // speed when animating touch move

    // unique element id for this slider
    this.uniqueId = (0|Math.random()*9e6).toString(36); // e.g. "1p7ym"

    this.slideList = rootElement.getElementsByClassName(slideListClass).item(0);
    this.slideItems = this.arrayFrom(rootElement.getElementsByClassName(slideItemClass));

    if (!this.slideList || !this.slideItems) return;

    this.slideItemMedia = this.arrayFrom(rootElement.getElementsByClassName(slideMediaClass));
    this.slideItemLinks = this.arrayFrom(rootElement.getElementsByClassName(slideMediaLinkClass));
    this.slideItemInfos = this.arrayFrom(rootElement.getElementsByClassName(slideMediaInfoClass));

    // target of the left/right button aria-controls
    this.slideListId = `c-photo-slider-${this.uniqueId}`;
    this.slideList.setAttribute("id", this.slideListId);

    // get all slide images (avoid grabbing any additional images [ads?])
    this.slideItemImages = [];
    for (let media of this.slideItemMedia) {
      media.classList.add(slideMediaImageFullFitClass);
      this.slideItemImages.push(media.getElementsByClassName(slideImageClass).item(0));
    }

    // grab all images, we don't want any ghosting for any of them
    this.slideItemGhosts = this.arrayFrom(rootElement.getElementsByTagName("img"));
    window.requestAnimationFrame(() => {
      // prevent all images from showing ghosts when dragged via mouse
      this.slideItemGhosts.map(element => {
        element.setAttribute("draggable", "false");
      });
      this.updateSlideStates();
    });

    // Fix this issue for Chrome
    // https://github.com/aFarkas/lazysizes/issues/137
    window.addEventListener("lazybeforesizes", function(e) {
      if (e.defaultPrevented) { return; }
		  var element = e.target;
      element.style.width = e.detail.width + "px";
    });

    // show/hide controls when an image is clicked
    if (this.slideItemLinks) {
      for (let i = 0; i < this.slideItemLinks.length; i++) {
        let slideItemId = this.slideListId + "-" + (i + 1);
        this.slideItemLinks[i].setAttribute("aria-role", "button");
        this.slideItemLinks[i].setAttribute("aria-label", "Toggle section");

        // for related information overlay
        if (this.slideItemInfos[i]) {
          this.slideItemLinks[i].setAttribute("aria-expanded", "true");
          this.slideItemLinks[i].setAttribute("aria-controls", slideItemId);
          this.slideItemInfos[i].setAttribute("id", slideItemId);
        }
      }
    }

    // previous slide button
    this.slideNavLeft = rootElement.getElementsByClassName(slideNavLeftClass).item(0);
    if (this.slideNavLeft) {
      this.slideNavLeft.setAttribute("aria-controls", this.slideListId);

      this.slideNavLeft.addEventListener("keydown", e => {
        if (this.isKeySpaceOrEnter(e) && this.hasNavShown()) {
          e.preventDefault();
          e.stopPropagation();
          this.slideTo(this.slideActive - 1, false, false); // previous
        }
      });

      this.slideNavLeft.addEventListener("click", e => {
        if (this.hasNavShown()) {
          e.preventDefault();
          e.stopPropagation();
          this.slideTo(this.slideActive - 1, false, false); // previous
        }
      });
    }

    // next slide button
    this.slideNavRight = rootElement.getElementsByClassName(slideNavRightClass).item(0);
    if (this.slideNavRight) {
      this.slideNavRight.setAttribute("aria-controls", this.slideListId);

      this.slideNavRight.addEventListener("keydown", e => {
        if (this.isKeySpaceOrEnter(e) && this.hasNavShown()) {
          e.preventDefault();
          e.stopPropagation();
          this.slideTo(this.slideActive + 1, false, false); // next
        }
      });

      this.slideNavRight.addEventListener("click", e => {
        if (this.hasNavShown()) {
          e.preventDefault();
          e.stopPropagation();
          this.slideTo(this.slideActive + 1, false, false); // next
        }
      });
    }

    // hide navigation arrows if only a single slide present
    if (this.slideItems.length == 1) {
      this.hideNav();
      return; // you shall not pass
    }

    // default slide
    this.slideTo(this.slideActive, true, false);

    window.requestAnimationFrame(() => {
      this.updateButtonStates();
    });

    // handle touch events
    this.initSwipes();

    // delegate the slide element click event to root element
    rootElement.addEventListener("click", e => {
      e.preventDefault();
      // ignore the left/right disabled buttons
      if (typeof e.target.disabled == "undefined") {
        this.toggleInterface();
      } else if (e.target.disabled === false) {
        this.toggleInterface();
      }
    });

    // keypress events for navigation
    rootElement.addEventListener("keydown", e => {
      let match = false;
      switch(e.keyCode) {
        case 34: // page down
        case 39: // right
        case 40: // down
          match = true;
          this.slideTo(this.slideActive + 1, true, false); // next
          break;

        case 33: // page up
        case 37: // left
        case 38: // up
          match = true;
          this.slideTo(this.slideActive - 1, true, false); // previous
          break;

        case 36: // home
          match = true;
          this.slideTo(0, true, false); // first
          break;

        case 35: // end
          match = true;
          this.slideTo(this.slideItems.length - 1, true, false); // last
          break;

        // delegate the slide element events to root element
        case 32: // spacebar
        case 13: // enter
          match = true;
          this.toggleInterface();
          break;
      }
      // allow all other key events bubble
      if (match && !e.metaKey) {
        e.preventDefault();
        e.stopPropagation();
      }
    });
  },

  toggleInterface: function() {
    window.requestAnimationFrame(() => {
      if (this.hasInfoShown()) {
        this.hideNav();
        this.hideInfo();
      } else {
        this.showNav();
        this.showInfo();
      }
    });
  },

  hasInfoShown: function() {
    return this.rootElement.classList.contains(slideHasInfoClass);
  },

  hasNavShown: function() {
    return this.rootElement.classList.contains(slideHasNavClass);
  },

  showNav: function() {
    // only show navigation if more than one slide present
    if (this.slideItems.length > 1) {
      this.rootElement.classList.add(slideHasNavClass);
    }
  },

  hideNav: function() {
    this.rootElement.classList.remove(slideHasNavClass);
  },

  showInfo: function() {
    this.rootElement.classList.add(slideHasInfoClass);
    this.slideItemLinks.map(element => {
      element.setAttribute("aria-expanded", "true");
    });
    this.slideItemInfos.map(element => {
      element.removeAttribute("aria-hidden");
    });
    ee.emitEvent(`${eventNamespace}:toggleinfo`, [true]);
  },

  hideInfo: function() {
    this.rootElement.classList.remove(slideHasInfoClass);
    this.slideItemLinks.map(element => {
      element.setAttribute("aria-expanded", "false");
    });
    this.slideItemInfos.map(element => {
      element.removeAttribute("aria-hidden");
    });
    ee.emitEvent(`${eventNamespace}:toggleinfo`, [false]);
  },

  slideTo: function(number, setFocus, animate) {
    // validate requested slide index
    if (number < 0) {
      this.slideActive = 0; // stop at first item (do not loop to last item)
    } else if (number > this.slideItems.length - 1) {
      this.slideActive = this.slideItems.length - 1; // stop at last item (do not loop to first item)
    } else {
      this.slideActive = number;
    }

    // only show slower animation for touch input, non-touch input is instant
    let animateSpeed = animate ? this.slideSpeed : 0;

    // setup the slide animate move
    window.requestAnimationFrame(() => {
      this.translateContent(this.slideActive, 0, animateSpeed);
      this.updateButtonStates();
      this.updateSlideStates();

      if (this.slideActive !== number) {
        // a setTimeout is needed here, since changing the style transition/transform above,
        // will not trigger a transitionend event on the element.
        reqTimeout.setTimeout(() => {
          // only set focus when event did not originate from mouse click
          // so set focus for touch input and key input
          if (setFocus && this.slideItemLinks[this.slideActive]) {
            this.slideItemLinks[this.slideActive].focus();
          }
          ee.emitEvent(`${eventNamespace}:selected`, [this.slideActive]);
        }, animateSpeed);
      }
    });
  },

  updateSlideStates: function() {
    // hide all slides, and prevent tab focus on link
    this.slideItemLinks.map(element => {
      element.setAttribute("tabindex", "-1");
    });
    this.slideItems.map(element => {
      element.setAttribute("aria-hidden", "true");
    });
    // show the active slide, and allow tab focus on link
    if (this.slideItemLinks[this.slideActive]) {
      this.slideItemLinks[this.slideActive].removeAttribute("tabindex");
    }
    if (this.slideItems[this.slideActive]) {
      this.slideItems[this.slideActive].removeAttribute("aria-hidden");
    }
  },

  updateButtonStates: function() {
    this.slideNavLeft.removeAttribute("aria-disabled");
    this.slideNavLeft.removeAttribute("disabled");

    this.slideNavRight.removeAttribute("aria-disabled");
    this.slideNavRight.removeAttribute("disabled");

    if (this.slideActive == 0) {
      this.slideNavLeft.setAttribute("aria-disabled", "true");
      this.slideNavLeft.setAttribute("disabled", "disabled");

    } else if (this.slideActive == this.slideItems.length - 1) {
      this.slideNavRight.setAttribute("aria-disabled", "true");
      this.slideNavRight.setAttribute("disabled", "disabled");
    }
  },

  isKeySpaceOrEnter: function(e) {
    let match = false;
    switch(e.keyCode) {
      case 32: // spacebar
      case 13: // enter
        if (!e.metaKey) {
          return true;
        }
        break;
    }
    return false;
  },

  arrayFrom: function(HTMLCollection) {
    if (!Array.from) {
      return (HTMLCollection ? [].slice.call(HTMLCollection) : []);
    } else {
      return Array.from(HTMLCollection);
    }
  },

  initSwipes: function() {
    let threshold = 0; // calc when to move to other slide
    let origX, origY; // original position of touch start event x/y axis
    let distX, distY; // distance travelled of touch move x/y axis

    // keep slider container width updated
    window.addEventListener("windowResize.throttled", () => {
      let newWidth = this.dimensions(this.rootElement).width;
      if (this.swipeContainerWidth !== newWidth) {
        this.swipeContainerWidth = newWidth;
        this.slideTo(this.slideActive, false, false); // go to current immediately
      }
    });
    this.swipeContainerWidth = this.dimensions(this.rootElement).width;

    this.rootElement.addEventListener("touchstart", e => {
      threshold = this.swipeContainerWidth * this.swipeThreshold;
      origX = e.touches[0].pageX;
      distX = 0;
      origY = e.touches[0].pageY;
      distY = 0;
    });

    this.rootElement.addEventListener("touchmove", e => {
      e.preventDefault(); // block browser zoom and scroll
      if (e.touches.length > 1 || e.scale && e.scale !== 1) {
        return;
      }
      distX = e.touches[0].pageX - origX;
      distY = e.touches[0].pageY - origY;

      window.requestAnimationFrame(() => {
        this.translateContent(this.slideActive, distX, this.touchSpeed);
      });
    });

    this.rootElement.addEventListener("touchend", e => {
      let directionX = 0;
      let directionY = 0;

      if (Math.abs(distX) > threshold) {
        directionX = distX > threshold ? 1 : -1;
      }
      distX = 0;

      if (Math.abs(distY) > threshold) {
        directionY = distY > threshold ? 1 : -1;
      }
      distY = 0;

      if (directionY === -1) {
        ee.emitEvent(`${eventNamespace}:swipeup`, [this.slideActive]); // close
      }

      if (directionX === 1) {
        if (this.slideActive > 0) {
          this.slideTo(this.slideActive - 1, true, true); // previous
        } else {
          this.slideTo(this.slideActive, true, true); // current
        }
      } else if (directionX === -1) {
        if (this.slideActive < this.slideItems.length - 1) {
          this.slideTo(this.slideActive + 1, true, true); // next
        } else {
          this.slideTo(this.slideActive, true, true); // current
        }
      } else {
        this.slideTo(this.slideActive, true, true); // current
      }
    });
  },

  translateContent: function(slideActive, offsetX, duration) {
    let slideX = -1 * slideActive * this.swipeContainerWidth;
    let positionX = (slideX + offsetX);

    this.slideList.style.webkitTransitionDuration = `${duration}ms`;
    this.slideList.style.mozTransitionDuration = `${duration}ms`;
    this.slideList.style.msTransitionDuration = `${duration}ms`;
    this.slideList.style.transitionDuration = `${duration}ms`;

    this.slideList.style.webkitTransform = `translate(${positionX}px,0) translateZ(0)`;
    this.slideList.style.mozTransform = `translate(${positionX}px,0)`;
    this.slideList.style.msTransform = `translate(${positionX}px,0)`;
    this.slideList.style.transform = `translate(${positionX}px,0) translateZ(0)`;
  },

  dimensions: function (el) {
    const rect = el.getBoundingClientRect();
    const width = rect.right - rect.left;
    const height = rect.bottom - rect.top;
    const bottom = rect.bottom;
    const top = rect.top;
    const ret = {
      height,
      width,
      top,
      bottom
    };
    return ret;
  }

};
